<?php
// Copyright (C) <2015>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

class Autoreport extends AutoreportModuleAppModel {

    public $hasAndBelongsToMany = [
        'User' => [
            'joinTable' => 'users_to_autoreports',
            'unique'    => true,
        ],
    ];

    var $actsAs = ['DateRange'];

    var $hasMany = [
        'AutoreportHostMembership'    => [
            'className' => 'AutoreportModule.AutoreportHostMembership',
            'dependent' => true,
        ],
        'AutoreportServiceMembership' => [
            'className' => 'AutoreportModule.AutoreportServiceMembership',
            'dependent' => true,
        ],
        'AutoreportAvailabilityLog'   => [
            'className' => 'AutoreportModule.AutoreportAvailabilityLog',
            'dependent' => true,
        ]
    ];

    var $belongsTo = [
        'Container'  => [
            'className' => 'Container',
            'dependent' => true,
        ],
        'Timeperiod' => [
            'className' => 'Timeperiod',
            'dependent' => true,
        ],
        'Calendar',
    ];

    public $validate = [
        'container_id'         => [
            'notBlank' => [
                'rule'     => ['notZero', 'container_id'],
                'message'  => 'This field cannot be left blank.',
                'required' => true,
            ],
        ],
        'name'                 => [
            'allowEmpty' => [
                'rule'     => 'notBlank',
                'message'  => 'This field cannot be left blank',
                'required' => true,
            ],
        ],
        'timeperiod_id'        => [
            'notBlank'       => [
                'rule'     => ['notZero', 'timeperiod_id'],
                'message'  => 'This field cannot be left blank.',
                'required' => true,
            ],
            'timeRangeValid' => [
                'rule'     => ['timeRangeValid', 'timeperiod_id'],
                'message'  => 'There are no time frames defined.Time evaluation report data is not available for the selected period.',
                'required' => true,
            ],
        ],
        'report_interval'      => [
            'notBlank' => [
                'rule'     => ['notZero', 'report_interval'],
                'message'  => 'This field cannot be left blank.',
                'required' => true,
            ],
        ],
        'report_send_interval' => [
            'notBlank' => [
                'rule'     => ['notZero', 'report_send_interval'],
                'message'  => 'This field cannot be left blank',
                'required' => true,
            ],
        ],
        'min_availability_on'  => [
            'atLeastOne'        => [
                'rule'       => ['atLeastOne'],
                'message'    => 'You must specify at least "Min. allowed availability" or "Max. number of outages"',
                'required'   => false,
                'allowEmpty' => true,
            ],
            'checkPercentValue' => [
                'rule'       => ['validatePercentValue'],
                'message'    => 'Please enter a number between 0 and 100 for "Min. allowed availability" in %%',
                'required'   => false,
                'allowEmpty' => true,
            ],
        ],
        'report_user'          => [
            'notBlank' => [
                'rule'    => ['validateUsers'],
                'message' => 'You must assign at least one User',
            ],
        ],
        'report_start_date'    => [
            'notBlank' => [
                'rule'    => ['validateStartSettings'],
                'message' => 'Start date cannot be left blank',
            ],
            'date'     => [
                'rule'       => ['date', 'ymd'],
                'message'    => 'Enter a valid date',
                'allowEmpty' => true,
            ],
        ],
        'calendar_id'          => [
            'notBlank' => [
                'rule'       => ['validateCalendarSettings'],
                'message'    => 'This field cannot be left blank',
                'allowEmpty' => true,
            ],
        ],
    ];

    public $validateStepTwo = [
        'host_id'        => [
            'notBlank' => [
                'allowEmpty' => false,
                'rule'       => [
                    'multiple', [
                        'min' => 1,
                    ]
                ],
                'message'    => 'You have to choose at least one host',
                'required'   => true,
            ],
        ],
        'start_validate' => [
            'checkData' => [
                'rule'     => ['validateHostServiceData'],
                'required' => true,
                'message'  => 'You must select at least one option for host and/or service ("in %%" and/or "in h")',
            ],
        ],
    ];


    public $validateCreateReport = [
        'autoreport_id' => [
            'notBlank' => [
                'allowEmpty' => false,
                'rule'       => [
                    'multiple', [
                        'min' => 1,
                    ]
                ],
                'message'    => 'This field cannot be left blank',
                'required'   => true,
            ],
        ],
        'start_date'    => [
            'notBlank'      => [
                'rule'     => 'notBlank',
                'required' => true,
                'message'  => 'This field cannot be left blank',
            ],
            'date'          => [
                'rule'     => ['date', 'dmy'],
                'required' => true,
                'message'  => 'Enter a valid date',
            ],
            'validateDates' => [
                'rule'     => ['validateDates', ['start_date', 'end_date']],
                'required' => true,
                'message'  => '"To" must not be earlier than "From"',
            ],
        ],
        'end_date'      => [
            'notBlank'      => [
                'rule'     => 'notBlank',
                'required' => true,
                'message'  => 'This field cannot be left blank',
            ],
            'date'          => [
                'rule'     => ['date', 'dmy'],
                'required' => true,
                'message'  => 'Enter a valid date',
            ],
            'validateDates' => [
                'rule'     => ['validateDates', ['start_date', 'end_date']],
                'required' => true,
                'message'  => '"To" must not be earlier than "From"',
            ],
        ],
    ];

    function __construct($id = false, $table = null, $ds = null) {
        parent::__construct($id, $table, $ds);
        $this->setValidationRules();
    }

    /*
    Custom validation rule for "Min. allowed availability" and/or "Max. number of outages" fields
    */
    public function atLeastOne($data) {
        $availability_array = Set::extract($this->data, $this->name . '.{(min_availability|max_.*outages).*}');

        //minimal availability in percent/minutes or max number of outages must be selected
        return (($availability_array['min_availability_on'] && ($availability_array['min_availability_in_percent'] || $availability_array['min_availability_in_minutes']))
            || ($availability_array['max_outages'] && $availability_array['max_number_of_outages']));
    }

    /*
    Custom validation rule for "Min. allowed availability" and/or "Max. number of outages" fields
    */
    public function validatePercentValue($data) {
        return (!($this->data['Autoreport']['min_availability_in_percent'] > 100));
    }

    public function notZero($field_name) {
        return !($this->data[$this->name][key($field_name)] == '0');
    }

    public function timeRangeValid($data) {
        ClassRegistry::init('Timeperiod');
        $selectedTimeperiod = $this->Timeperiod->findById($data['timeperiod_id']);

        return !(empty($selectedTimeperiod['Timerange']));
    }

    public function validateUsers($data) {
        return !($this->data[$this->name]['report_send_interval'] !== 'NEVER' && empty($data['report_user']));
    }

    public function validateStartSettings($data) {
        return !($this->data[$this->name]['use_start_time'] && empty($data['report_start_date']));
    }

    public function validateCalendarSettings($data) {
        return !($this->data[$this->name]['consider_holidays'] && empty($data['calendar_id']));
    }

    public function validateHostServiceData() {
        if (empty($this->data['Autoreport']['Host'])) {
            return false;
        }
        return !empty(Hash::filter(Hash::filter(Hash::merge(
            Set::classicExtract($this->data, 'Autoreport.Host.{n}.{(host_(?!outage|show_all_failures|alias)[^"]+)}'),
            Set::classicExtract($this->data, 'Autoreport.Host.{n}.Service.{n}.{(service_(?!outage|show_all_failures|graph)[^"]+)}')
        ), ['Autoreport', 'filterNullValues'])));
    }

    public function validateDates() {
        return (strtotime($this->data['Autoreport']['start_date']) <= strtotime($this->data['Autoreport']['end_date']));
    }

    public function prepareDataForView($host_and_services_arr, $saved_settings = []) {
        foreach ($host_and_services_arr as $key => $host_and_services) {
            if (!empty($host_and_services['Service'])) {
                foreach ($host_and_services['Service'] as $service_key => $host_and_service) {
                    //if servicetemplate has been deleted
                    if (sizeof($host_and_service['Servicetemplate']) === 0) {
                        unset($host_and_services['Service'][$service_key]);
                        continue;
                    }
                    $host_and_services['Service'][$service_key]['name'] = (empty($host_and_service['name'])) ?
                        $host_and_service['Servicetemplate']['name'] : $host_and_service['name'];
                    $host_and_services['Service'][$service_key]['process_performance_data'] = (is_null($host_and_service['process_performance_data'])) ?
                        $host_and_service['Servicetemplate']['process_performance_data'] : $host_and_service['process_performance_data'];
                    $name[$service_key] = $host_and_services['Service'][$service_key]['name'];

                }
                $host_and_services['Service'] = Hash::sort($host_and_services['Service'], '{n}.name', 'asc', 'string');
                $host_and_services_arr[$key] = $host_and_services;
            }
        }
        if (!empty($saved_settings)) {
            foreach ($host_and_services_arr as $key => $data_arr) {
                $saved_config = Hash::extract($saved_settings, 'AutoreportHostMembership.{n}[host_id=' . $data_arr['Host']['id'] . ']');
                $host_and_services_arr[$key]['Host'] = Hash::merge(
                    $host_and_services_arr[$key]['Host'],
                    (!empty($saved_config)) ? [
                        'outage_duration'      => $saved_config[0]['outage_duration'],
                        'configuration_option' => $saved_config[0]['configuration_option'],
                    ] : []
                );

                foreach ($data_arr['Service'] as $sub_key => $service_data) {
                    $saved_config = Hash::extract($saved_settings, 'AutoreportServiceMembership.{n}[service_id=' . $service_data['id'] . ']');
                    $host_and_services_arr[$key]['Service'][$sub_key] = Hash::merge(
                        $host_and_services_arr[$key]['Service'][$sub_key],
                        (!empty($saved_config)) ? [
                            'outage_duration'      => $saved_config[0]['outage_duration'],
                            'configuration_option' => $saved_config[0]['configuration_option'],
                            'graph_settings'       => $saved_config[0]['graph_settings'],
                        ] : []
                    );
                }
            }
        }

        return $host_and_services_arr;
    }

    public function prepareDataForSave($request_data, $autoreport_id = null, $saved_settings = []) {
        $data_array = [];
        $autoreportData = Hash::extract($request_data, 'Autoreport.Host');
        foreach ($autoreportData as $host_id => $value) {
            $validate_check = Hash::apply(Set::extract($value, '{(host_percent|host_minutes)}'), '{s}', 'array_sum');
            if ($validate_check) {
                $configuration_option = Hash::apply(Set::extract($value, '{(host_(?!outage))}'), '{s}', 'array_sum');
                if ($configuration_option > 0) {
                    $saved_config = Hash::extract($saved_settings, 'AutoreportHostMembership.{n}[host_id=' . $host_id . ']');
                    $data_array['AutoreportHostMembership'][] = [
                        'id'                   => (!empty($saved_config)) ? $saved_config[0]['id'] : null,
                        'autoreport_id'        => $autoreport_id,
                        'host_id'              => $host_id,
                        'configuration_option' => $configuration_option,
                        'outage_duration'      => (empty($value['host_outage'])) ? null : $value['host_outage'],
                    ];
                }
            }
            $autoreport_services = Hash::extract($value, 'Service');
            if (!empty($autoreport_services)) {
                foreach ($autoreport_services as $service_id => $service) {
                    $validate_check = Hash::apply(Set::extract($service, '{(service_percent|service_minutes)}'), '{s}', 'array_sum');
                    if ($validate_check) {
                        $configuration_option = Hash::apply(Set::extract($service, '{(service_(?!outage))}'), '{s}', 'array_sum');
                        if ($configuration_option > 0) {
                            $saved_config = Hash::extract($saved_settings, 'AutoreportServiceMembership.{n}[service_id=' . $service_id . ']');
                            $data_array['AutoreportServiceMembership'][] = [
                                'id'                   => (!empty($saved_config)) ? $saved_config[0]['id'] : null,
                                'autoreport_id'        => $autoreport_id,
                                'host_id'              => $host_id,
                                'service_id'           => $service_id,
                                'configuration_option' => $configuration_option,
                                'outage_duration'      => (empty($service['service_outage'])) ? null : $service['service_outage'],
                                'graph_settings'       => (empty($service['graph_settings'])) ? null : $service['graph_settings']
                            ];
                        }
                    }
                }
            }
        }
        return Hash::filter($data_array);
    }

    /**
     * Here we can check some conditions to see what validation we need
     * @return boolean
     */
    public function beforeValidate($options = []) {
        $router_params = Router::getParams();
        if (isset($router_params['action'])) {
            switch ($router_params['action']) {
                case 'addStepTwo':
                case 'editStepTwo':
                    $this->validate = $this->validateStepTwo;
                    break;
                case 'createReport':
                    $this->validate = $this->validateCreateReport;
                    break;
            }
        }

        return true;
    }


    /**
     * Callback function for filtering.
     *
     * @param array $var Array to filter.
     *
     * @return boolean
     */
    public static function filterNullValues($var) {
        if ($var != '0') {
            return true;
        }

        return false;
    }

    public function setValidationRules($condition = null) {
        if ($condition == 'createReport') {
            $this->validate = $this->validateCreateReport;
        }
    }

    public static function calculateTotalTime($time_slice) {
        return $time_slice['end'] - $time_slice['start'];
    }

    /**
     * Generate date for host/service objects
     *
     * @param integer $total_time total time of all time ranges
     * @param array $time_slices => Time ranges for evaluation period
     * @param array $state_history_with_object => Array with object data and state history array
     * @param boolean $check_hard_state use hard and soft state or only hard state for evaluation
     * @param boolean $object_host , if object host => 1, else service => 0
     * @param         array                      downtimes array, use downtimes array if report setting
     *                                                     "consider_downtimes" => true
     *
     * @return array $autoreport_data for host/service object
     */
    public function generateAutoreportData($time_slices, $state_history_with_object, $check_hard_state, $object_host = false) {
        $state_array = array_fill(0, ($object_host) ? 3 : 4, 0); // host states => 0,1,2; service statea => 0,1,2,3
        $state_ok = 0;
        $state_unknown = ($object_host) ? 2 : 3;//if the end of date in the future
        $evaluation_data = Hash::merge($state_array, [
            'outages' => [],
        ]);
        $state_history_array = Hash::extract($state_history_with_object, 'Statehistory.{n}');
        $outage_state = ($object_host) ? 1 : 2; // 1 for host down and 2 for service critical
        $set_initial_state = false;
        $current_state = 0;
        $outage_counter = 0;
        foreach ($time_slices as $time_slice_key => $time_slice) {
            $time = $time_slice['start'];
            if ($time > strtotime('today 23:59:59')) { // ignore time_slice in the future
                $current_state = $state_unknown;
            }
            $is_downtime = $time_slice['is_downtime'];
            reset($state_history_array);
            foreach ($state_history_array as $key => $state_history) {
                $state_time_timestamp = $state_history['state_time'];
                if (!$set_initial_state) {
                    $current_state = $state_history['last_state'];
                    if ($check_hard_state && $state_history['last_state'] != 0) {
                        $current_state = $state_history['last_hard_state'];
                    }
                    $current_state = ($current_state == -1) ? 0 : $current_state;
                    $set_initial_state = true;
                }
                if ($state_time_timestamp >= $time_slice['end']) {
                    // if state time after time slice
                    break;
                }
                if ($state_time_timestamp <= $time_slice['start']) {
                    $current_state = ($state_history['state'] == 0 || !$check_hard_state || ($check_hard_state && ($check_hard_state && $state_history['is_hardstate']))) ? $state_history['state'] : $current_state;
                    unset($state_history_array[$key]);
                    continue;
                }
                if ($state_time_timestamp > $time_slice['start']) {
                    //if outage in downtime add time for state "ok"
                    if (($current_state == $outage_state) && $is_downtime) {
                        $evaluation_data[0] += $state_time_timestamp - $time;
                        //$current_state = $state_ok;
                        $evaluation_data['outages'][$outage_counter] = [
                            'start'       => $time,
                            'end'         => $state_time_timestamp,
                            'is_downtime' => $is_downtime,
                        ];
                        $outage_counter++;
                    } else {
                        $evaluation_data[$current_state] += $state_time_timestamp - $time;
                        if ($current_state == $outage_state) {
                            $evaluation_data['outages'][$outage_counter] = [
                                'start'       => $time,
                                'end'         => $state_time_timestamp,
                                'is_downtime' => $is_downtime,
                            ];
                            $outage_counter++;
                        }
                    }
                    $current_state = ($state_history['state'] == 0 || !$check_hard_state || ($check_hard_state && ($check_hard_state && $state_history['is_hardstate']))) ? $state_history['state'] : $current_state;
                    $time = $state_time_timestamp;
                    unset($state_history_array[$key]);
                }
            }
            //if outage in downtime add time for state "ok"
            if ($current_state == $outage_state && $is_downtime) {
                $evaluation_data[$state_ok] += $time_slice['end'] - $time;
                //$current_state = $state_ok;
                if ($current_state == $outage_state) {
                    $evaluation_data['outages'][$outage_counter] = [
                        'start'       => $time,
                        'end'         => $time_slice['end'],
                        'is_downtime' => $is_downtime,
                    ];
                    $outage_counter++;
                }
            } else {
                $evaluation_data[$current_state] += $time_slice['end'] - $time;

                if ($current_state == $outage_state) {
                    $evaluation_data['outages'][$outage_counter] = [
                        'start'       => $time,
                        'end'         => $time_slice['end'],
                        'is_downtime' => $is_downtime,
                    ];
                    $outage_counter++;
                }
            }
        }
        $evaluation_data['outages_summarized_in_downtime'] = $this->mergeOutages(
            Hash::extract($evaluation_data, 'outages.{n}[is_downtime=1]')
        );
        $evaluation_data['outages'] = $this->mergeOutages(
            Hash::extract($evaluation_data, 'outages.{n}[is_downtime=0]')
        );
        if (!empty($evaluation_data['outages'])) {
            $evaluation_data['maximum_outage_duration'] = Hash::apply(Hash::map($evaluation_data['outages'], '{n}', ['Autoreport', 'calculateTotalTime']), '{n}', 'max');
        }
        unset($time_slices, $state_history_with_object);
        return $evaluation_data;
    }

    public function mergeOutages($outage_array) {

        $new_outages_arr = [];
        $tmp_outageEnd = 0;
        $tmp_outageStart = 0;

        $array_counter = 0;
        foreach ($outage_array as $key => $outage) {
            $set_end = false;
            $outageStart = $outage['start'];
            $outageEnd = $outage['end'];
            if ($tmp_outageStart == 0 && $tmp_outageEnd == 0) {
                $tmp_outageStart = $outageStart;
                $tmp_outageEnd = $outageEnd;
                $new_outages_arr[$array_counter] = ['start' => $outageStart, 'end' => $outageEnd];
                continue;
            }
            if ($outageStart >= $tmp_outageStart && $outageEnd <= $tmp_outageEnd) {
                continue;
            } else if ($outageStart >= $tmp_outageStart && $outageStart < $tmp_outageEnd && $outageEnd > $tmp_outageEnd) {
                $tmp_outageEnd = $outageEnd;
                $new_outages_arr[$array_counter] = ['start' => $tmp_outageStart, 'end' => $outageEnd];
            } else if ($outageStart > $tmp_outageStart && $outageStart > $tmp_outageEnd) {
                $array_counter++;
                $tmp_outageStart = $outageStart;
                $tmp_outageEnd = $outageEnd;
                $new_outages_arr[$array_counter] = ['start' => $outageStart, 'end' => $outageEnd];
            }
        }

        return $new_outages_arr;
    }

    /*Cronjob functions*/
    public function hasToBeSend($last_send_date, $report_send_interval) {
        $now = time();
        $has_to_be_send = false;

        if (strcmp($report_send_interval, 'NEVER') != 0) {
            if ($last_send_date == '0000-00-00 00:00:00') {
                return true;
            }
            $last_send_timestamp = strtotime($last_send_date);
            if (strcmp($report_send_interval, 'DAY') == 0) {
                if (intval(date('Ymd', $now)) > intval(date('Ymd', $last_send_timestamp))) {
                    $has_to_be_send = true;
                }
            } else if (strcmp($report_send_interval, 'WEEK') == 0) {
                if (intval(date('oW', $now)) > intval(date('oW', $last_send_timestamp))) {
                    $has_to_be_send = true;
                }
            } else if (strcmp($report_send_interval, 'MONTH') == 0) {
                if (intval(date('Ym', $now)) > intval(date('Ym', $last_send_timestamp))) {
                    $has_to_be_send = true;
                }
            } else if (strcmp($report_send_interval, 'QUARTER') == 0) {
                $currentMonth = date('n', $now);
                $currentQuarter = ceil($currentMonth / 3);
                $lastSendMonth = date('n', $last_send_timestamp);
                $lastSendQuarter = ceil($lastSendMonth / 3);
                if (intval(date('Y', $now) . $currentQuarter) > intval(date('Y', $last_send_timestamp) . $lastSendQuarter)) {
                    $has_to_be_send = true;
                }
            } else if (strcmp($report_send_interval, 'YEAR') == 0) {
                if (intval(date('Y', $now)) > intval(date('Y', $last_send_timestamp))) {
                    $has_to_be_send = true;
                }
            }
        }

        return $has_to_be_send;
    }

    public function reportStartTime($report_interval) {
        $now = $this->reportEndTime($report_interval);
        $dateNow = new DateTime(date('d.m.Y H:i:s', $now));
        switch ($report_interval) {
            case 'WEEK':
                $dateNow->modify('last Monday');
                break;
            case 'MONTH':
                $dateNow->modify('first day of this month');
                break;
            case 'QUARTER':
                $currentMonth = date('n');
                $currentYear = date('o');
                if ($currentMonth >= 1 && $currentMonth <= 3) {
                    $dateNow->modify('first day of october ' . ($currentYear - 1));
                } else if ($currentMonth >= 4 && $currentMonth <= 6) {
                    $dateNow->modify('first day of january ' . $currentYear);
                } else if ($currentMonth >= 7 && $currentMonth <= 9) {
                    $dateNow->modify('first day of april ' . $currentYear);
                } else if ($currentMonth >= 10 && $currentMonth <= 12) {
                    $dateNow->modify('first day of july ' . $currentYear);
                }
                break;
            case 'YEAR':
                $dateNow->modify('first day of january ' . $dateNow->format('Y'));
                break;
        }
        $dateNow->setTime(0, 0, 0);
        return $dateNow->getTimestamp();
    }

    public function reportCurrentStartTime($report_interval) {
        $dateNow = new DateTime(date('d.m.Y H:i:s', time()));
        switch ($report_interval) {
            case 'WEEK':
                $dateNow->modify('last Monday');
                break;
            case 'MONTH':
                $dateNow->modify('first day of this month');
                break;
            case 'QUARTER':
                $currentMonth = date('n');
                $currentYear = date('o');
                if ($currentMonth >= 1 && $currentMonth <= 3) {
                    $dateNow->modify('first day of january' . ($currentYear - 1));
                } else if ($currentMonth >= 4 && $currentMonth <= 6) {
                    $dateNow->modify('first day of april ' . $currentYear);
                } else if ($currentMonth >= 7 && $currentMonth <= 9) {
                    $dateNow->modify('first day of july ' . $currentYear);
                } else if ($currentMonth >= 10 && $currentMonth <= 12) {
                    $dateNow->modify('first day of october ' . $currentYear);
                }
                break;
            case 'YEAR':
                $dateNow->modify('first day of january ' . $dateNow->format('Y'));
                break;
        }
        $dateNow->setTime(0, 0, 0);
        return $dateNow->getTimestamp();
    }

    public function reportInitStartTime($report_interval) {
        $dateNow = new DateTime(date('d.m.Y H:i:s', time()));
        switch ($report_interval) {
            case 'WEEK':
                $dateNow->modify('first monday of january ' . $dateNow->format('Y'));
                break;
            case 'DAY':
            case 'MONTH':
            case 'QUARTER':
            case 'YEAR':
                $dateNow->modify('first day of january ' . $dateNow->format('Y'));
                break;
        }
        $dateNow->setTime(0, 0, 0);
        return $dateNow->getTimestamp();
    }

    public function reportEndTime($report_interval) {
        $dateNow = new DateTime(date('d.m.Y H:i', time()));
        switch ($report_interval) {
            case 'DAY':
                $dateNow->modify('yesterday');
                break;
            case 'WEEK':
                $dateNow->modify('last Sunday');
                break;
            case 'MONTH':
            case 'YEAR':
                $dateNow->modify('last day of last month');
                break;
            case 'QUARTER':
                $currentMonth = date('n');
                $currentYear = date('o');
                if ($currentMonth >= 1 && $currentMonth <= 3) {
                    $dateNow->modify('last day of december ' . ($currentYear - 1));
                } else if ($currentMonth >= 4 && $currentMonth <= 6) {
                    $dateNow->modify('last day of march ' . $currentYear);
                } else if ($currentMonth >= 7 && $currentMonth <= 9) {
                    $dateNow->modify('last day of june ' . $currentYear);
                } else if ($currentMonth >= 10 && $currentMonth <= 12) {
                    $dateNow->modify('last day of september ' . $currentYear);
                }
                break;
        }

        $dateNow->setTime(23, 59, 59);
        return $dateNow->getTimestamp();
    }

    /**
     * array structure have too look like the following example:
     * array(
     *  'hostIds' => array(
     *      (int) 0 => (int) 1,
     *      (int) 1 => (int) 37
     *  ),
     *  'serviceIds' => array(
     *      (int) 0 => (int) 2,
     *      (int) 1 => (int) 3,
     *      (int) 2 => (int) 4,
     *      (int) 3 => (int) 1,
     *      (int) 4 => (int) 160,
     *      (int) 5 => (int) 161,
     *      (int) 6 => (int) 162,
     *      (int) 7 => (int) 163
     *  )
     * )
     * @param $ids - new host and service ids or already assigned ids from the report
     * @param $diffIds - ids which are removed from the report
     */
    public function setUsageFlags($ids, $diffIds = []) {
        $this->Host = ClassRegistry::init('Host');
        $this->Service = ClassRegistry::init('Service');
        //check usage flags
        foreach ($ids['hostIds'] as $hostId) {
            $hasFlag = $this->Host->checkUsageFlag($hostId, AUTOREPORT_MODULE);
            if (!$hasFlag && empty($diffIds['hostIds'])) {
                $newFlag = $this->Host->currentUsageFlag + AUTOREPORT_MODULE;
                $this->Host->id = $hostId;
                //update field usage_flag
                $this->Host->saveField('usage_flag', $newFlag);
            } else {
                if ($hasFlag && !empty($diffIds['hostIds']) && in_array($hostId, $diffIds['hostIds'])) {
                    if ($this->checkForAdditionalUse($hostId, 'Host') == 0) {
                        //the host is not in use by any other Report so delete the flag
                        $newFlag = $this->Host->currentUsageFlag - AUTOREPORT_MODULE;
                        $this->Host->id = $hostId;
                        $this->Host->saveField('usage_flag', $newFlag);
                    }
                }
            }
        }

        foreach ($ids['serviceIds'] as $serviceId) {
            $hasFlag = $this->Service->checkUsageFlag($serviceId, AUTOREPORT_MODULE);
            if (!$hasFlag && empty($diffIds['serviceIds'])) {
                $newFlag = $this->Service->currentUsageFlag + AUTOREPORT_MODULE;
                $this->Service->id = $serviceId;
                //update field usage_flag
                $this->Service->saveField('usage_flag', $newFlag);
            } else {
                if ($hasFlag && !empty($diffIds['serviceIds']) && in_array($serviceId, $diffIds['serviceIds'])) {
                    if ($this->checkForAdditionalUse($serviceId, 'Service') == 0) {
                        //the service is not in use by any other Report so delete the flag
                        $newFlag = $this->Service->currentUsageFlag - AUTOREPORT_MODULE;
                        $this->Service->id = $serviceId;
                        $this->Service->saveField('usage_flag', $newFlag);
                    }
                }
            }
        }
    }

    /**
     * checks if the given element is in ue by another autoreport
     * @param $id - ID to lookup
     * @param $type - host or service to lookup
     * @return mixed
     */
    public function checkForAdditionalUse($id, $type) {
        $this->AutoreportHostMembership = ClassRegistry::init('AutoreportHostMembership');
        $this->AutoreportServiceMembership = ClassRegistry::init('AutoreportServiceMembership');

        if ($type == 'Host') {
            $count = $this->AutoreportHostMembership->find('count', [
                'recursive'  => -1,
                'conditions' => [
                    'host_id' => $id
                ]
            ]);
            return $count;
        }

        if ($type == 'Service') {
            $count = $this->AutoreportServiceMembership->find('count', [
                'recursive'  => -1,
                'conditions' => [
                    'service_id' => $id
                ]
            ]);
            return $count;
        }
    }

    public function getHosts() {
        $result = $this->AutoreportHostMembership->find('all', [
            'recursive' => -1,
            'fields'    => [
                'DISTINCT host_id'
            ]
        ]);
        $result = Hash::extract($result, '{n}.AutoreportHostMembership.host_id');
        return $result;
    }

    public function getServices() {
        $result = $this->AutoreportServiceMembership->find('all', [
            'recursive' => -1,
            'fields'    => [
                'DISTINCT service_id'
            ]
        ]);
        $result = Hash::extract($result, '{n}.AutoreportServiceMembership.service_id');
        return $result;
    }

    /**
     * @param string $key
     * @param array $downtimes
     * @param array $systemfailures
     * @return array
     */
    public function mergeDowntimesWithSystemfailures($downtimes = [], $systemfailures = []) {
        debug($downtimes);
        debug($systemfailures);

        return Hash::sort($downtimes, '{n}.scheduled_start_time', 'ASC');
    }

    /**
     * @param $autoreport
     * @return array date ranges with [start => timestamp, end => timestamp]
     */
    public function getTimeRangesForAutoreport($autoreport) {
        if (!empty($autoreport['AutoreportAvailabilityLog']['last_update'])) {
            $initStartDate = $autoreport['AutoreportAvailabilityLog']['last_update'];
        } else {
            $initStartDate = $this->reportInitStartTime($autoreport['Autoreport']['report_interval']);
        }
        $dateRanges = $this->createDateSlicesByIntervalAndLastUpdateDate(
            $initStartDate,
            $autoreport['Autoreport']['report_interval']
        );

        return $dateRanges;
    }
}
