<?php
// Copyright (C) <2015>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, version 3 of the License.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//	If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//	under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//	License agreement and license key will be shipped with the order
//	confirmation.

/**
 * Class MkInstallShell
 * uses CakePHP to insert MK Template data into the Database
 *
 * usage: oitc mk_module.mk_install
 */
class MkInstallShell extends AppShell {
    public $uses = [
        'Servicetemplate',
        'Hosttemplate',
        'Contact',
        'Command',
        'Commandargument',
        'Mkchecks',
    ];
    public $mapping = [];
    //root path for json and check files
    public $path = '';
    public $contactId = [1];

    public function main() {
        $this->stdout->styles('success', ['text' => 'green']);
        $this->out('Installing MK Module');
        $this->path = APP . 'Plugin' . DS . 'MkModule' . DS . 'webroot' . DS;
        $jsonDataPath = 'jsonData' . DS;
        try {
            $this->checkDependencies();
            $iterator = new DirectoryIterator($this->path . $jsonDataPath);
            $files = [];
            foreach ($iterator as $file) {
                if (!$file->isDot()) {
                    $filename = $file->getFilename();
                    if (preg_match('/.+_command_data.json/', $filename)) {
                        $key = 'command';
                    } else if (preg_match('/.+_servicetemplate_data.json/', $filename)) {
                        $key = 'servicetemplate';
                    } else if (preg_match('/.+_mkchecks_data.json/', $filename)) {
                        $key = 'mkchecks';
                    } else {
                        continue;
                    }
                    $files[$key] = $file->getFilename();
                }
            }
        } catch (Exception $e) {
            $this->hr(1);
            $this->error($e->getMessage());
        }
        if (empty($files['command'])) {
            $this->error('no Commands found');
        }

        if (empty($files['servicetemplate'])) {
            $this->error('no Template data found');
        }
        //save the data into the DB
        //commands
        $commandData = $this->readJsonFile($this->path . $jsonDataPath . $files['command']);
        $this->saveData($commandData);

        //servicetemplates
        $stData = $this->readJsonFile($this->path . $jsonDataPath . $files['servicetemplate']);
        $servicetemplateData = $this->modifyServicetemplatedata($stData);
        $this->saveData($servicetemplateData);

        //mkchecks
        $stData = $this->readJsonFile($this->path . $jsonDataPath . $files['mkchecks']);
        $mkchecksData = $this->modifyMkChecksdata($stData);
        $this->saveData($mkchecksData);
    }

    /**
     * check if dependencies for installing the templates
     * @throws Exception
     */
    public function checkDependencies() {
        $this->out('<info>Checking dependencies for installing Templates</info>');
        $contact = $this->Contact->find('first', [
            'recursive'  => -1,
            'conditions' => [
                'Contact.name' => 'info'
            ],
            'fields'     => [
                'Contact.id'
            ]
        ]);
        if (empty($contact)) {
            throw new Exception('Found no "info" contact on your System. To Continue the Installation Please create an "info" Contact!');
        }
        $this->contactId = Hash::extract($contact, 'Contact.id');
        $this->out('<success>done!</success>');
    }

    /**
     * saves the Template Data into the Database
     *
     * @param $dataToSave - the array data to save
     *
     * @return bool
     */
    private function saveData($dataToSave) {
        try {
            $this->out('Saving data to DB');
            $model = key($dataToSave[0]);
            foreach ($dataToSave as $data) {
                //calling model->create() so we can use save in a loop
                $this->{$model}->create();
                //check if the record already exists
                $existingData = $this->{$model}->find('first', [
                    'recursive'  => -1,
                    'conditions' => [
                        $model . '.name' => $data[$model]['name']
                    ]
                ]);
                if (!empty($existingData)) {
                    //the data already exists. Skip the loop for this record
                    $this->out('<info>' . $data[$model]['name'] . ' already exists. Skipping</info>');
                    switch ($model) {
                        case 'Command':
                            //add mapping for commands with the already existing record from the database
                            $this->mapping['Command'][$data['Command']['uuid']] = $existingData['Command']['id'];
                            $this->mapCommandArgs($existingData['Command']['id']);
                            break;
                    }
                    continue;
                }
                if ($model == 'Command') {
                    $data = Hash::remove($data, 'Commandargument.{n}.id');
                }
                if ($this->{$model}->saveAll($data)) {
                    switch ($model) {
                        case 'Command':
                            $this->out('<info>Add Command ' . $data['Command']['name'] . ' to Mapping</info>');
                            $this->mapping['Command'][$data['Command']['uuid']] = $this->{$model}->id;
                            $this->mapCommandArgs($this->{$model}->id);
                            break;
                    }
                    $this->out('<success>Data for ' . $model . ' ' . $data[$model]['name'] . ' sucessfully saved!</success>');
                } else {
                    print_r($data);
                    print_r($this->{$model}->validationErrors);
                    throw new Exception('the data for ' . $model . ' ' . $data[$model]['name'] . ' could not be saved');
                }
            }
        } catch (Exception $e) {
            echo $e->getMessage();
            echo "\n";
        }
        return false;
    }

    /**
     * reads a json file and return the json representation as php array
     *
     * @param $file - the json file to read
     *
     * @return array - the php array representation of the json
     */
    private function readJsonFile($file) {
        try {
            $this->out('Reading json file ' . $file);
            if (!file_exists($file)) {
                throw new Exception($file . ' could not be found');
            }
            $fileData = file_get_contents($file);
            $data = json_decode($fileData, true);
            $jsonErr = json_last_error();
            if ($jsonErr != JSON_ERROR_NONE) {
                throw  new Exception($jsonErr);
            }
            return $data;
        } catch (Exception $e) {
            $this->error($e->getMessage());
        }
    }

    /**
     * @param $servicetemplateData
     *
     * @return array
     */
    private function modifyServicetemplatedata($servicetemplateData) {
        $this->out('<info>Replacing Servicetemplate IDs </info>');
        if (!empty($this->mapping['Commandarguments'])) {
            foreach ($servicetemplateData as $key => $servicetemplate) {
                $commandId = $this->mapping['Command'][$servicetemplate['Servicetemplate']['command_id']];
                //replacing the command uuid with the command id
                $servicetemplateData[$key]['Servicetemplate']['command_id'] = $commandId;
                $i = 0;
                //go through the servicetemplatecommandarguments
                if (!empty($servicetemplate['Servicetemplatecommandargumentvalue'])) {
                    foreach ($servicetemplate['Servicetemplatecommandargumentvalue'] as $stcavKey => $stcav) {
                        //replace the commandargument ids from the servicetemplatecommandarguments
                        // with the commandargument ids (the id they have after save)
                        if (array_key_exists($commandId, $this->mapping['Commandarguments']) && !empty($this->mapping['Commandarguments'][$commandId])) {
                            $servicetemplateData[$key]['Servicetemplatecommandargumentvalue'][$stcavKey]['commandargument_id'] = $this->mapping['Commandarguments'][$commandId][$i];
                            $i++;
                        }
                    }
                }
            }
        } else {
            $this->out('<error>There is no mapping for Commandarguments so the Servicetemplates cant be Saved!</error>');
        }
        return $servicetemplateData;
    }

    /**
     * @param $mkchecksData
     *
     * @return array
     */
    private function modifyMkChecksdata($mkChecksData) {
        $this->out('<info>Mapping Servicetemplate to MK Checks </info>');
        foreach ($mkChecksData as $key => $mkCheck) {
            //debug($mkCheck['Mkchecks']['servicetemplate_id']);
            $templateToMap = $this->Servicetemplate->find('all', [
                'recursive'  => -1,
                'conditions' => [
                    [
                        'Servicetemplate.uuid' => $mkCheck['Mkchecks']['servicetemplate_id']
                    ]
                ],
                'fields'     => [
                    'id',
                ]
            ]);
            $mkChecksData[$key]['Mkchecks']['servicetemplate_id'] = $templateToMap[0]['Servicetemplate']['id'];
        }
        return $mkChecksData;
    }

    /**
     * maps the command ids with the commandargument ids
     *
     * @param $commandId
     *
     * @return void
     */
    private function mapCommandArgs($commandId) {
        $commands = $this->Command->find('all', [
            'conditions' => [
                'id' => $commandId,
            ],
        ]);
        $commandargIds = Hash::filter(Hash::extract($commands, '{n}.Commandargument.{n}.id'));
        $this->mapping['Commandarguments'][$commandId] = $commandargIds;
    }
}