L.HexbinLayer = L.Layer.extend({
    _undef: function _undef(a) {
        return typeof a === 'undefined';
    },

    options: {
        radius: 25,
        opacity: 0.6,
        duration: 200,
        valueDomain: [-1, 0, 1, 2, 3],
        colorRange: ['#003EF9', '#00C838', '#FFD938', '#E92223', '#808080'],

        onmouseover: function onmouseover(d) {
            console.log("mouseover");
        },
        onmouseout: undefined,
        click: function click(d) {
            console.log("click");
        },
        lng: function lng(d) {
            return d.longitude;
        },
        lat: function lat(d) {
            return d.latitude;
        },
        value: function value(d) {
            //return _.meanBy(d, (o) => o.o.colorcode)
            var stats = [];
            d.forEach(function (o) {
                stats.push(o.o.colorcode);
            });
            return stats.reduce(function (a, b) {
                return Math.max(a, b);
            }, -Infinity);
        }
    },

    initialize: function initialize(options) {
        L.setOptions(this, options);

        this._data = [];
        this._colorScale = d3.scale.linear().domain(this.options.valueDomain).range(this.options.colorRange).clamp(true);
    },
    onAdd: function onAdd(map) {
        this.map = map;
        var _layer = this;

        // SVG element
        this._svg = L.svg();
        map.addLayer(this._svg);
        this._rootGroup = d3.select(this._svg._rootGroup).classed('d3-overlay', true);
        this.selection = this._rootGroup;

        // Init shift/scale invariance helper values
        this._pixelOrigin = map.getPixelOrigin();
        this._wgsOrigin = L.latLng([0, 0]);
        this._wgsInitialShift = this.map.latLngToLayerPoint(this._wgsOrigin);
        this._zoom = this.map.getZoom();
        this._shift = L.point(0, 0);
        this._scale = 1;

        // Create projection object
        this.projection = {
            latLngToLayerPoint: function latLngToLayerPoint(latLng, zoom) {
                zoom = _layer._undef(zoom) ? _layer._zoom : zoom;
                var projectedPoint = _layer.map.project(L.latLng(latLng), zoom)._round();
                return projectedPoint._subtract(_layer._pixelOrigin);
            },
            layerPointToLatLng: function layerPointToLatLng(point, zoom) {
                zoom = _layer._undef(zoom) ? _layer._zoom : zoom;
                var projectedPoint = L.point(point).add(_layer._pixelOrigin);
                return _layer.map.unproject(projectedPoint, zoom);
            },
            unitsPerMeter: 256 * Math.pow(2, _layer._zoom) / 40075017,
            map: _layer.map,
            layer: _layer,
            scale: 1
        };
        this.projection._projectPoint = function (x, y) {
            var point = _layer.projection.latLngToLayerPoint(new L.LatLng(y, x));
            this.stream.point(point.x, point.y);
        };

        this.projection.pathFromGeojson = d3.geo.path().projection(d3.geo.transform({ point: this.projection._projectPoint }));

        // Compatibility with v.1
        this.projection.latLngToLayerFloatPoint = this.projection.latLngToLayerPoint;
        this.projection.getZoom = this.map.getZoom.bind(this.map);
        this.projection.getBounds = this.map.getBounds.bind(this.map);
        this.selection = this._rootGroup; // ???

        // Initial draw
        this.draw();
    },
    onRemove: function onRemove(map) {
        if (this._container != null) this._container.remove();

        // Remove events
        map.off({ 'moveend': this._redraw }, this);

        this._container = null;
        this._map = null;

        // Explicitly will leave the data array alone in case the layer will be shown again
        // this._data = [];
    },
    addTo: function addTo(map) {
        map.addLayer(this);
        return this;
    },
    _disableLeafletRounding: function _disableLeafletRounding() {
        this._leaflet_round = L.Point.prototype._round;
        L.Point.prototype._round = function () {
            return this;
        };
    },
    _enableLeafletRounding: function _enableLeafletRounding() {
        L.Point.prototype._round = this._leaflet_round;
    },
    draw: function draw() {
        this._disableLeafletRounding();
        this._redraw(this.selection, this.projection, this.map.getZoom());
        this._enableLeafletRounding();
    },

    getEvents: function getEvents() {
        return { zoomend: this._zoomChange };
    },
    _zoomChange: function _zoomChange(evt) {
        this._disableLeafletRounding();
        var newZoom = this._undef(evt.zoom) ? this.map._zoom : evt.zoom; // "viewreset" event in Leaflet has not zoom/center parameters like zoomanim
        this._zoomDiff = newZoom - this._zoom;
        this._scale = Math.pow(2, this._zoomDiff);
        this.projection.scale = this._scale;
        this._shift = this.map.latLngToLayerPoint(this._wgsOrigin)._subtract(this._wgsInitialShift.multiplyBy(this._scale));

        var shift = ["translate(", this._shift.x, ",", this._shift.y, ") "];
        var scale = ["scale(", this._scale, ",", this._scale, ") "];
        this._rootGroup.attr("transform", shift.concat(scale).join(""));

        this.draw();
        this._enableLeafletRounding();
    },
    _redraw: function _redraw(selection, projection, zoom) {
        var _this = this;

        // Generate the mapped version of the data
        var data = this._data.map(function (d) {
            var lng = _this.options.lng(d);
            var lat = _this.options.lat(d);

            var point = projection.latLngToLayerPoint([lat, lng]);
            return { o: d, point: point };
        });

        // Select the hex group for the current zoom level. This has
        // the effect of recreating the group if the zoom level has changed
        var join = selection.selectAll('g.hexbin').data([zoom], function (d) {
            return d;
        });

        // enter
        join.enter().append('g').attr('class', function (d) {
            return 'hexbin zoom-' + d;
        });

        // exit
        join.exit().remove();

        d3.selectAll('path.hexbin-hexagon').remove();

        // add the hexagons to the select
        this._createHexagons(join, data, projection);
    },
    _createHexagons: function _createHexagons(g, data, projection) {
        var _this2 = this;

        // Create the bins using the hexbin layout
        var hexbin = d3.hexbin().radius(this.options.radius / projection.scale).x(function (d) {
            return d.point.x;
        }).y(function (d) {
            return d.point.y;
        });
        var bins = hexbin(data);

        // Join - Join the Hexagons to the data
        var join = g.selectAll('path.hexbin-hexagon').data(bins);

        // Update - set the fill and opacity on a transition (opacity is re-applied in case the enter transition was cancelled)
        join.transition().duration(this.options.duration).attr('fill', function (d) {
            return _this2._colorScale(_this2.options.value(d));
        }).attr('fill-opacity', this.options.opacity).attr('stroke-opacity', this.options.opacity);

        // Enter - establish the path, the fill, and the initial opacity
        join.enter().append('path').attr('class', 'hexbin-hexagon').attr('d', function (d) {
            return 'M' + d.x + ',' + d.y + hexbin.hexagon();
        }).attr('fill', function (d) {
            return _this2._colorScale(_this2.options.value(d));
        }).attr('fill-opacity', 0.01).attr('stroke-opacity', 0.01).on('mouseover', this.options.mouseover).on('mouseout', this.options.mouseout).on('click', this.options.click).transition().duration(this.options.duration).attr('fill-opacity', this.options.opacity).attr('stroke-opacity', 0.00);

        // Exit
        join.exit().transition().duration(this.options.duration).attr('fill-opacity', 0.01).attr('stroke-opacity', 0.01).remove();
    },
    data: function data(_data) {
        this._data = _data != null ? _data : [];
        this.draw();
        return this;
    }
});