<?php
// Copyright (C) <2015>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, version 3 of the License.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

// 2.
//	If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//	under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//	License agreement and license key will be shipped with the order
//	confirmation.


/**
 * Class SAPDataShell
 * uses CakePHP to get SAP Template and command data from itcockpit database
 */
App::uses('Folder', 'Utility');
App::uses('File', 'Utility');

class SAPDataShell extends AppShell {

    public $uses = [
        'Servicetemplate',
        'Command',
        'Commandargument',
        'Timeperiod',
    ];

    /** @var array - maps the command_id to a uuid
     *  this is needed for the new command_ids when the commands are saved into a database
     */
    public $mapping = [];

    public $timeperiodId = [1];

    public $servicetemplateData = [];

    public $commandData = [];

    public function main() {
        if ($this->checkNecessaryData()) {
            $this->servicetemplateData = $this->getSapServicetemplateData();
            $this->commandData = $this->getCommandData();
            // $this->mapData();
            $this->out('<info>Start writing Files</info>');
            $this->writeFilesAsJson('sap_servicetemplates', $this->servicetemplateData);
            $this->writeFilesAsJson('sap_commands', $this->commandData);
        } else {
            $this->out('<error>Necessary data was not found</error>');
        }
    }

    /**
     * check if the necessary data exists on the system. In this case the must be the Timeperiod with ID 1 or the
     * name 24/7.
     * @return bool
     */
    public function checkNecessaryData() {
        //check if check_period_id and notify_period_id (Timeperiod) 1 exists (should be 24/7)

        $idExists = $this->Timeperiod->exists($this->timeperiodId[0]);

        if (!$idExists) {
            //@TODO test this!!
            $nameExists = $this->Timeperiod->find('first', [
                'recursive'  => -1,
                'conditions' => [
                    'Timeperiod.name LIKE' => '%24/7%',
                ]
            ]);
            $this->timeperiodID = Hash::extract($nameExists, 'Timeperiod.id');
            if (empty($nameExists)) {
                return false;
            }
            return true;
        }
        return $idExists;
    }


    /**
     * Get all command data which keys exists in the mapping
     * @return array
     */
    public function getCommandData() {
        //get the needed command keys
        $this->out('<info>Get command data</info>');
        $commandKeys = array_keys($this->mapping);
        $commands = $this->Command->find('all', [
            'recursive'  => -1,
            'conditions' => [
                'id' => $commandKeys
            ],
            'fields'     => [
                'id',
                'name',
                'command_line',
                'command_type',
                'human_args',
                'uuid',
                'description',
            ],
            'contain'    => [
                'Commandargument' => [
                    'fields' => [
                        // 'id',
                        // 'command_id',
                        'name',
                        'human_name'
                    ]
                ]
            ]
        ]);

        //replace the uuid of the command with the uuid from mapping
        foreach ($commands as $key => $command) {
            $commands[$key]['Command']['uuid'] = $this->mapping[$command['Command']['id']];
        }

        //remove useless ids
        $commands = Hash::remove($commands, '{n}.Command.id');
        $commands = Hash::remove($commands, '{n}.Commandargument.{n}.command_id');

        return $commands;
    }

    /**
     * @param string $filter - name filter from user input @TODO not implemented yet
     * @return array
     */
    public function getSapServicetemplateData($filter = '') {
        $this->out('<info>Get servicetemplate data</info>');
        $st = $this->Servicetemplate->find('all', [
            'recursive'  => -1,
            'conditions' => [
                'OR' => [
                    ['Servicetemplate.name LIKE' => 'sap%'],
                    ['Servicetemplate.name LIKE' => 'RFC-Destination-NAME%']
                ]
            ],
            'fields'     => [
                'id',
                'name',
                'template_name',
                'servicetemplatetype_id',
                'description',
                'command_id', // must be overridden with uuid and also with the same uuid for uuid in command data
                'check_command_args',
                'checkcommand_info',
                'check_interval',
                'retry_interval',
                'max_check_attempts',
                'first_notification_delay',
                'notification_interval',
                'notify_on_warning',
                'notify_on_unknown',
                'notify_on_critical',
                'notify_on_recovery',
                'notify_on_flapping',
                'notify_on_downtime',
                'flap_detection_enabled',
                'flap_detection_on_ok',
                'flap_detection_on_warning',
                'flap_detection_on_unknown',
                'flap_detection_on_critical',
                'low_flap_threshold',
                'high_flap_threshold',
                'process_performance_data',
                'freshness_checks_enabled',
                'freshness_threshold',
                'passive_checks_enabled',
                'event_handler_enabled',
                'active_checks_enabled',
                'retain_status_information',
                'retain_nonstatus_information',
                'notifications_enabled',
                'notes',
                'priority',
                'tags',
                'service_url',
                'is_volatile',
                'check_freshness',
            ],
            'contain'    => [
                'Contact',
                'Contactgroup',
                'Servicetemplatecommandargumentvalue' => [
                    'fields' => [
                        //'commandargument_id',
                        'value',
                        //'servicetemplate_id'
                    ]
                ]
            ]
        ]);

        //remove the useless ids
        $st = Hash::remove($st, '{n}.Servicetemplate.id');
        $st = Hash::remove($st, '{n}.Servicetemplatecommandargumentvalue.{n}.servicetemplate_id');

        //replace command_id with an uuid which also will inserted as command UUID when we get the command data
        foreach ($st as $key => $servicetemplate) {
            if (isset($this->mapping[$servicetemplate['Servicetemplate']['command_id']])) {
                //command id has already an uuid in mapping
                $st[$key]['Servicetemplate']['command_id'] = $this->mapping[$servicetemplate['Servicetemplate']['command_id']];
            } else {
                //no mapping for this command exists so get a new uuid
                $currentUuid = UUID::v4();
                $this->mapping[$servicetemplate['Servicetemplate']['command_id']] = $currentUuid;
                $st[$key]['Servicetemplate']['command_id'] = $currentUuid;
            }

            //container id is 1 every time
            $st[$key]['Servicetemplate']['container_id'] = 1;
            //eventhandler_command_id is 0 every time
            $st[$key]['Servicetemplate']['eventhandler_command_id'] = 0;
            //check_period_id and notify_period is always the 24/7 timeperiod
            $st[$key]['Servicetemplate']['check_period_id'] = $this->timeperiodId[0];
            $st[$key]['Servicetemplate']['notify_period_id'] = $this->timeperiodId[0];

        }

        return $st;
    }


    public function writeFilesAsJson($name, $content) {
        if (empty($content)) {
            $this->out('<error>The ' . $name . ' File could not be written because its empty</error>');
            return false;
        }
        $dir = APP . 'Plugin' . DS . $this->plugin . DS . 'webroot' . DS;
        $folder = 'jsonData';
        if (!is_dir($dir . $folder)) {
            mkdir($dir . $folder, 0755);
        }
        $filename = $name . '.json';
        $file = new File($dir . $folder . DS . $filename, true, 0644);
        if (!$file->exists()) {
            $file->create();
        }
        $content = json_encode($content);
        $return = $file->write($content);
        $file->close();

        if ($return) {
            $this->out('<success>File ' . $filename . ' was sucessfully written to ' . $dir . $folder . '</success>');
        } else {
            $this->out('<error>The File ' . $filename . ' could not be written</error>');
        }
        return $return;
    }
}
