#!/usr/bin/perl
# --
# check_sap_background_jobs.pl - Nagios Plugion to monitor SAP Background Jobs
# Copyright (C) 2013-2014 it-novum GmbH, http://www.it-novum.com
# --
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
# --

use strict;
use warnings;

use FindBin qw($Bin);
use lib "$Bin";

# lib fallback for Module::SAPRfc
unless ( $ENV{LD_LIBRARY_PATH} ) {
    $ENV{LD_LIBRARY_PATH} = "/usr/sap/rfcsdk/lib";
}

use Config::Simple;
use Date::Calc;
use Module::SAPRfc;
use Nagios::Plugin;

use constant CONFIGFILE => $FindBin::Bin . '/sap_config.ini';
use vars qw($VERSION);

$VERSION = '0.2.0';

=head1 CHANGELOG

0.2.0 [27.01.2014]  - Added parameter -S <ConfigSection> to use different sap systems
                    - Improoved error handling using _die
                    - Append all jobs found to the check output

0.1.0 [11.12.2013]  - Initial release

=over 4

=cut

# ---------------------------------- #
# parameterization
# ---------------------------------- #

my $Usage = "This is $0 Version $VERSION\n\n"
    . "Usage: %s -S <SAPConfigSectionName> -m <JobsOfLastXMinutesFilter> -j <JobnameFilter> "
    . "-s <StatusFilter> -w <Warning> -c <Critical>\n"
    . "Uses connection attributes defined in config file " . CONFIGFILE;

my $NagiosObject = Nagios::Plugin->new(
    usage => $Usage,
);

$NagiosObject->add_arg(
    spec => 'ConfigSection|S=s',
    help => '-S, --ConfigSection=STRING (Sektion name of one connection defined in sap_config.ini)',
);

# TODO: allow only integer instead of strings
$NagiosObject->add_arg(
    spec => 'LastMinutes|m=s',
    help => '-m, --LastMinutes=INTEGER (jobs started in the last x minutes)',
);

$NagiosObject->add_arg(
    spec    => 'Jobname|j=s',
    default => '*',
    help    => '-j, --Jobname=STRING',
);

$NagiosObject->add_arg(
    spec => 'Status|s=s',
    help => '-s, --Status=STRING (Prelim, Schedul, Ready, Running, Finished, Aborted)',
);

$NagiosObject->add_arg(
    spec    => 'warning|w=s',
    default => '0:',
    help    => '-w, --warning=STRING',
);

$NagiosObject->add_arg(
    spec    => 'critical|c=s',
    default => '0:',
    help    => '-c, --critical=STRING',
);

$NagiosObject->getopts;
my $Thresholds = $NagiosObject->set_thresholds(
    warning  => $NagiosObject->opts->warning,
    critical => $NagiosObject->opts->critical,
);

# ---------------------------------- #
# config
# ---------------------------------- #

my $ConfigObject = Config::Simple->new();
$ConfigObject->read(CONFIGFILE) or _die( $ConfigObject->error() || '' );

my $Config = $ConfigObject->param(
    -block => $NagiosObject->opts->ConfigSection,
);

if ( ref $Config ne 'HASH' || !%{$Config} ) {
    _die( 'Given section (-S) not found in ' . CONFIGFILE );
}

# ---------------------------------- #
# core acions
# ---------------------------------- #

my %Filter;
my $SAPRfcObject = Module::SAPRfc->new(
    Host         => $Config->{Host},
    Client       => $Config->{Client},
    SystemNumber => $Config->{SystemNumber},
    User         => $Config->{User},
    Password     => $Config->{Password},
    SystemID     => $Config->{SystemID},
);

if ( ref $SAPRfcObject ne 'Module::SAPRfc' ) {
    _die('Could not instantiate SAPRfcObject!');
}

# complete status filter
if ( $NagiosObject->opts->Status ) {

    # validate
    my @StatusValid = qw( Prelim Schedul Ready Running Finished Aborted );
    unless ( grep { $NagiosObject->opts->Status eq $_ } @StatusValid ) {

        _die('Status invalid!');
    }

    $Filter{ $NagiosObject->opts->Status } = 'X';
}
if ( $NagiosObject->opts->LastMinutes ) {

    if ( $NagiosObject->opts->LastMinutes !~ m/^([0-9]+)$/ ) {
        _die('Given last minutes (-m) invalid. Only integers are allowed!');
    }

    my ( $sec, $min, $hour, $mday, $month, $year, $wday, $yday, $isdst ) = localtime(time);
    $year += 1900;

    $Filter{ToDate} = sprintf( "%04i%02i%02i", $year, ( $month + 1 ), $mday );
    $Filter{ToTime} = sprintf( "%02i%02i%02i", $hour, $min, $sec );

    ( $year, $month, $mday, $hour, $min, $sec ) = Date::Calc::Add_Delta_DHMS(
        $year, $month + 1, $mday, $hour, $min, $sec,
        0, 0, ( int( $NagiosObject->opts->LastMinutes ) * -1 ), 0
    );

    $Filter{FromDate} = sprintf( "%04i%02i%02i", $year, ($month), $mday );
    $Filter{FromTime} = sprintf( "%02i%02i%02i", $hour, $min, $sec );
}

# login
$SAPRfcObject->Call(
    Function  => 'BAPI_XMI_LOGON',
    Interface => 'XBP',
    Parameter => {
        Version => '1.0',
    },
);

# retrieve job count via job select
my $CallResult = $SAPRfcObject->Call(
    Function         => 'BAPI_XBP_JOB_SELECT',
    ExternalUserName => '*',
    JobSelectParam   => {
        Jobname => ( $NagiosObject->opts->Jobname || '*' ),
        Username => '*',
        %Filter,
    },
    Return => [qw( JobHead )],
);
my $JobCount     = scalar( @{ $CallResult->{JobHead} } );
my $NagiosOutput = sprintf(
    'There are %s Background Jobs with Pattern %s',
    $JobCount, $NagiosObject->opts->Jobname,
);

# append job names to the output
my @JobnameList
    = map { my $Job = $_->{JOBNAME}; $Job =~ s/([\s]+)$//; $Job } @{ $CallResult->{JobHead} };
for ( my $i = 1; $i <= scalar @JobnameList; $i++ ) {
    $NagiosOutput .= "\nJob $i: " . $JobnameList[ $i - 1 ];
}

# handle nagios exit
my $NagiosExitCode = $NagiosObject->check_threshold($JobCount);
$NagiosObject->nagios_exit( $NagiosExitCode, $NagiosOutput );

=begin Internal:

=cut

sub _die {
    $NagiosObject->nagios_die( shift(@_) );
}

=end Internal:

=back

=head1 TERMS AND CONDITIONS

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

=cut
