#!/usr/bin/perl -w
# perl script to call the webservice of SAP diagnostics agents
# --
# Copyright (C) 2013 it-novum GmbH, http://www.it-novum.com
# --
# Authors: Maximilian Grosch, Jens Michelsons
# --
# License ...
# --


#
# Patched for openITCOCKPIT V3 !!!!
# This script using oitc.cmd not nagios checkresults spool dir!!!!!!!
#

use strict;
use warnings;

use SOAP::Lite;
use Data::Dumper;
use Nagios::Plugin;
use Config::Simple;
use File::Temp;
use File::Touch;


use constant {
    STATE_OK => 0,
    STATE_WARNING => 1,
    STATE_CRITICAL => 2,
    STATE_UNKNOWN => 3,
};

=head1 NAME

Check Diagnostics Agents

=head1 SYNOPSIS

Check for SAP Diagnostics Agents

=head1 PUBLIC INTERFACE

=over 4

=cut

my $NagiosBasePath = '/opt/openitc/nagios';
my $CheckResultsDir = $NagiosBasePath . '/var/spool/checkresults/';

################################
# instanciate Nagios::Plugin   #
################################

my %Objects;
my $PluginName = "check_sap_diagnostics_agents.pl";

$Objects{NagiosPlugin} = Nagios::Plugin->new(
    usage => "
GetAlertTree:
    Single: $PluginName -H <hostname> -m GetAlertTree -p <port> -e <element>
    Multiple: $PluginName -H GetAlertTree -m GetAlertTree -p <port> -f <path_to_configuration>

ABAPGetWPTable: $PluginName -H <hostname> -m ABAPGetWPTable -W <workprocess> -p <port>
"
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'hostname|H=s',
    required => 1,
    help     => '-H, --hostname=STRING',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'ipaddress|I=s',
    help     => '-I, --ipaddress=STRING (optional if you cannot lookup hostname)',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'mode|m=s',
    required => 1,
    help     => '-m, --mode=STRING (ABAPGetWPTable,GetAlertTree)',
);

$Objects{NagiosPlugin}->add_arg(
    spec => 'element|e=s',
    help => '-e, --element=STRING (element to check)',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'port|p=i',
    required => 1,
    help     => '-p, --port=INTEGER (default: 5XX13, XX=Systemnummer)',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'workprocess|W=s',
    default  => 'UPD',
    help     => '-W, --workprocess=STRING; define workprocess (Default: UPD)',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'status|s=s',
    default  => 'Stopped',
    help     => '-s, --status=STRING; define critical state of workprocess (Default: Stopped)',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'warning|w=s',
    default  => '0:',
    help     => '-w, --warning=STRING',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'critical|c=s',
    default  => '0:',
    help     => '-c, --critical=STRING',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'configfile|f=s',
    help     => '-f, --configfile=STRING',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'list|l=s',
    help     => '-l, --list=STRING',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'ssl|S',
    help     => '-S, --ssl',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'soapuser|U=s',
    help     => '-U, --soapuser',
);

$Objects{NagiosPlugin}->add_arg(
    spec     => 'soappassword|P=s',
    help     => '-P, --soappassword',
);

$Objects{NagiosPlugin}->getopts;

my $thresholds = $Objects{NagiosPlugin}->set_thresholds(
    warning => $Objects{NagiosPlugin}->opts->warning,
    critical => $Objects{NagiosPlugin}->opts->critical
);

################################
# load configuration           #
################################

$Objects{Config} = '';
my %Config;

if (defined($Objects{NagiosPlugin}->opts->configfile)) {
    $Objects{Config} = new Config::Simple($Objects{NagiosPlugin}->opts->configfile);
    %Config = $Objects{Config}->vars();
}

################################
# instanciate SOAP Object      #
################################
my $protcol = 'http';

if (defined($Objects{NagiosPlugin}->opts->ssl)) {
    $protcol = 'https';
}

my $ConnectionHost = $Objects{NagiosPlugin}->opts->hostname;

if (defined($Objects{NagiosPlugin}->opts->ipaddress)) {
    $ConnectionHost = $Objects{NagiosPlugin}->opts->ipaddress;
}

$Objects{SOAP} = SOAP::Lite->new(
    proxy => $protcol . '://' . $ConnectionHost . ':' .
    $Objects{NagiosPlugin}->opts->port
);

$Objects{SOAP}->default_ns('urn:SAPControl');

################################
# main operations              #
################################

# ABAPGetWPTable
if ($Objects{NagiosPlugin}->opts->mode eq "ABAPGetWPTable") {
    my @ResultABAPGetWPTable =_ABAPGetWPTable(
        WpType => $Objects{NagiosPlugin}->opts->workprocess,
        WpCritState => $Objects{NagiosPlugin}->opts->status
    );

    my @ExitList = _ParseExitABAPGetWPTable(
        Data => \@ResultABAPGetWPTable,
    );

    _NagiosExit(Data => \@ExitList);
}

# GetAlertTree
if ($Objects{NagiosPlugin}->opts->mode eq "GetAlertTree") {
    my @ResultGetAlertTree = _GetAlertTree();

    my @ExitList = _ParseExitGetAlertTree(
        Data => \@ResultGetAlertTree,
    );

    if (defined($Objects{NagiosPlugin}->opts->element)) {
        _NagiosExit(Data => \@ExitList);
    }

    if (defined($Objects{NagiosPlugin}->opts->configfile)) {
        _NagiosSubmitPassiveData(Data => \@ExitList);
    }
}

# EnqGetStatistic
if ($Objects{NagiosPlugin}->opts->mode eq "EnqGetStatistic") {
    my @ResultEnqGetStatistic = _EnqGetStatistic();

#    my @ExitList = _ParseExit(
#        Data => \@ResultEnqGetStatistic,
#    );
#
#    if (defined($Objects{NagiosPlugin}->opts->element)) {
#        _NagiosExit(Data => \@ExitList);
#    }
#
#    if (defined($Objects{NagiosPlugin}->opts->configfile)) {
#        _NagiosSubmitPassiveData(Data => \@ExitList);
#    }
}

# Override SOAP Authentication
sub SOAP::Transport::HTTP::Client::get_basic_credentials {
    return $Objects{NagiosPlugin}->opts->soapuser => $Objects{NagiosPlugin}->opts->soappassword;
}

################################
# get data from webservice     #
################################

=item _ABAPGetWPTable
to get a list of work processes filtered by WpType (dialog process, update process, ...)

    my @ResultSet = _ABAPGetWPTable(
        WpType => 'UPD',    (optional, e.g. UPD = UpdateProcess)
    );

Return:
    @ResultSet = (
        {
            Type   => 'DIA',
            Status => 'Wait',
        },
        {...},
    );

=cut

sub _ABAPGetWPTable {
    my %Param  = @_;
    my $Som    = $Objects{SOAP}->call("ABAPGetWPTable");
    my $Result = $Som->result;
    my @ResultSet;

    foreach my $RowNum (@{$Result->{item}}) {
        if ($RowNum->{Status} eq $Param{WpCritState} && $RowNum->{Typ} eq $Param{WpType}) {
            push @ResultSet,{
                Typ => $RowNum->{Typ},
                Status => $RowNum->{Status}
            };
        }
    }

    return @ResultSet;
}

=item _GetAlertTree
Returns CCMS Alert tree as an array.
The parent-child node relationship is encoded via the parent index of each node.

    my $.. = _GetAlertTree();

Return:
returns the ....

=cut

sub _GetAlertTree {
    my %Param  = @_;
    my $Som    = $Objects{SOAP}->call("GetAlertTree");
    my $Result = $Som->result;
    my @ResultSet;

    my $Elements = scalar(@{$Result->{item}});
    for(my $ElementID=0; $ElementID<$Elements; $ElementID++) {
        my $CurrentElementID = $ElementID;
        my $Path = '';
        my $ServiceNameCount = 0;
        my $ServiceName = '';

        do {
            if ($ServiceNameCount lt 2) {
                $ServiceName = '-' .
                    $ServiceName if($Path && $Result->{item}->[$CurrentElementID]->{name});
                $ServiceName = $Result->{item}->[$CurrentElementID]->{name} . $ServiceName;
                $ServiceNameCount++;
            }

            $Path = '::' . $Path if($Path && $Result->{item}->[$CurrentElementID]->{name});
            $Path = $Result->{item}->[$CurrentElementID]->{name} . $Path;
            $CurrentElementID = $Result->{item}->[$CurrentElementID]->{parent};
        } while( $CurrentElementID ne -1 );

        $Result->{item}->[$ElementID]->{path} = $Path;
        $Result->{item}->[$ElementID]->{ServiceName} = $ServiceName;

        if (defined($Objects{NagiosPlugin}->opts->list)) {
            my $RegExp = $Objects{NagiosPlugin}->opts->list;
            if ($Path =~ /$RegExp/i) {
                print $Path . "\r\n    Default-ServiceName => (" . $ServiceName . ")\r\n\r\n";
            }
        }
    }

    # if -l option is used exit
    if (defined($Objects{NagiosPlugin}->opts->list)) {
        exit STATE_UNKNOWN;
    }

    # check for the given element
    if (defined($Objects{NagiosPlugin}->opts->element)) {
        foreach my $RowNum (@{$Result->{item}}) {
            my $Element = $Objects{NagiosPlugin}->opts->element . '$';
	        if ($RowNum->{path} =~ $Element) {
		        push @ResultSet,{
		            Name => $RowNum->{name},
                    Description => $RowNum->{description},
                    Status => $RowNum->{ActualValue},
                    Tid => $RowNum->{Tid},
		        };
            }
        }
    }

    # get a set of values to process passive
    if (defined($Objects{NagiosPlugin}->opts->configfile)) {
        # read all configured Objects from configuration file and look them up
        # in information from diagnostics agent
        foreach my $ConfigElement (keys %{$Objects{Config}->param(-block=>'GetAlertTree')}) {
            foreach my $RowNum (@{$Result->{item}}) {
                # Extend Regexp by $ to check by end of string
                $ConfigElement = $ConfigElement . '$';

                # if ConfigElement matches push result to ResultSet
                if ($RowNum->{path} =~ $ConfigElement) {
                    $RowNum->{ServiceName} = _CheckServiceNameMapping( $RowNum->{ServiceName} );

                    push @ResultSet,{
                        Name => $RowNum->{name},
                        Description => $RowNum->{description},
                        Status => $RowNum->{ActualValue},
                        ServiceName => $RowNum->{ServiceName},
                        Tid => $RowNum->{Tid},
                    };
                }
            }
        }

    }

    if (!@ResultSet) {
        print "UNKNOWN - No information gathered";
        exit STATE_UNKNOWN;
    }

    return @ResultSet;
}

=item _EnqGetStatistic
get Enqueue Statistics from diagnostics agent

Function needs to be implemented
Return:

=cut

sub _GetAlerts {
    my $RootTid = $_[0];
    my $Som    = $Objects{SOAP}->call("GetAlerts",SOAP::Data->name('RootTid')->value($RootTid));
    my @Result = $Som->paramsout;
    my @ResultSet;


    my %ExitCodeList = (
        "SAPControl-GREEN" => STATE_OK,
        "SAPControl-YELLOW" => STATE_WARNING,
        "SAPControl-RED" => STATE_CRITICAL,
        "SAPControl-GRAY" => STATE_UNKNOWN,
    );

    if (ref($Result[0]) eq 'HASH') {
        if (ref($Result[0]->{item}) eq 'ARRAY') {
            foreach my $RowNum (@{$Result[0]->{item}}) {
                push @ResultSet,{
                    Time => $RowNum->{Time},
                    ExitCode => $ExitCodeList{$RowNum->{Value}},
                    Description => $RowNum->{Description},
                };
            }
        } else {
            push @ResultSet,{
                Time => $Result[0]->{item}->{Time},
                ExitCode => $ExitCodeList{$Result[0]->{item}->{Value}},
                Description => $Result[0]->{item}->{Description},
            };
        }
    }

    return @ResultSet;
}

=item _EnqGetStatistic
get Enqueue Statistics from diagnostics agent

Function needs to be implemented
Return:

=cut

sub _EnqGetStatistic {
    my %Param  = @_;
    my $Som    = $Objects{SOAP}->call("EnqGetStatistic");
    my @Result = $Som->paramsall;
    my @ResultSet;

    print Data::Dumper::Dumper @Result;
    #foreach my $RowNum (@{$Result->{item}}) {
    #    if ($RowNum->{Status} eq $Param{WpCritState} && $RowNum->{Typ} eq $Param{WpType}) {
    #        push @ResultSet,{
    #            Typ => $RowNum->{Typ},
    #            Status => $RowNum->{Status}
    #        };
    #    }
    #}
    exit STATE_WARNING;
    #return @ResultSet;
}

#################################
# parse data for exit           #
#################################

sub _ParseExitABAPGetWPTable {
    my %Param  = @_;
    my @NagiosExit;

    # If case monitoring ABAPGetWPTable
    if ($Objects{NagiosPlugin}->opts->mode eq "ABAPGetWPTable") {
        my $CountWPStates = 0;
        foreach my $WPEntry (@{$Param{Data}}) {
            if ($Objects{NagiosPlugin}->opts->status eq $WPEntry->{Status}) {
                $CountWPStates++;
            }
        }

        my $ExitCode = $Objects{NagiosPlugin}->check_threshold(
            check => $CountWPStates,
        );

        push @NagiosExit,{
            PerfValue => $CountWPStates,
            ExitCode => $ExitCode,
            Output => " There are " . $CountWPStates . " workprocesses of Typ " .
                $Objects{NagiosPlugin}->opts->workprocess . " in state " .
                $Objects{NagiosPlugin}->opts->status,
        }
    }
    return @NagiosExit;
}

sub _ParseExitGetAlertTree {
    my %Param  = @_;
    my @NagiosExit;
    my %ExitCodeList = (
        "SAPControl-GREEN" => STATE_OK,
        "SAPControl-YELLOW" => STATE_WARNING,
        "SAPControl-RED" => STATE_CRITICAL,
        "SAPControl-GRAY" => STATE_UNKNOWN,
    );
    my $PerfData;
    my $warning;
    my $critical;

    # If case monitoring GetAlertTree
    if ($Objects{NagiosPlugin}->opts->mode eq "GetAlertTree") {
        foreach my $AlertEntry (@{$Param{Data}}) {
            my $ExitCode = $ExitCodeList{$AlertEntry->{Status}} // STATE_UNKNOWN;

            my $ServiceLongOutput = '';
            my $NewLine;

            if ($Objects{NagiosPlugin}->opts->element) {
                $NewLine = "\n\r";
            } else {
                $NewLine = '\n';
            }
            my @Alerts = _GetAlerts($AlertEntry->{Tid});
            foreach my $Alert (@Alerts) {
                $ServiceLongOutput .= $NewLine . 'Time: ' . $Alert->{Time} . ' Description: ' .
                    $Alert->{Description};
            }

            # if there is a performance value process it
            if ($AlertEntry->{Description} =~ m/^([0-9]+)[\s]*([\/|%|\w]*)/) {
                if (defined($Objects{NagiosPlugin}->opts->configfile)) {
                    if (defined($Objects{Config}->param('GetAlertTreeThresholds.' .
                        $AlertEntry->{ServiceName}))) {
                        my @thresholds = split(/;/,$Objects{Config}->param(
                            'GetAlertTreeThresholds.' . $AlertEntry->{ServiceName}
                        ));

                        $warning = $thresholds[0];
                        $critical = $thresholds[1];

                    }
                } else {
                    $warning  = $Objects{NagiosPlugin}->opts->warning;
                    $critical = $Objects{NagiosPlugin}->opts->critical;
                }

                $ExitCode = $Objects{NagiosPlugin}->check_threshold(
                    check => $1,
                    warning => $warning,
                    critical => $critical,
                );

                push @NagiosExit,{
                    PerfValue => $1,
                    #$NagiosExit{PerfUOM} = $2;
                    ExitCode => $ExitCode,
                    Output => $AlertEntry->{Name} . " -> " . $AlertEntry->{Description} .
                        $ServiceLongOutput,
                    Name => $AlertEntry->{Name},
                    ServiceName => $AlertEntry->{ServiceName},
                    Warning => $warning,
                    Critical => $critical,
                }
            } else {
                push @NagiosExit,{
                    ExitCode => $ExitCode,
                    Output => $AlertEntry->{Name} . " -> " . $AlertEntry->{Description} .
                        $ServiceLongOutput,
                    Name => $AlertEntry->{Name},
                    ServiceName => $AlertEntry->{ServiceName},
                };
            }
        }
    }
    return @NagiosExit;
}

################################
# submit data to nagios        #
################################

sub _NagiosExit{
    my %Param  = @_;

    if (defined($Param{Data}[0]->{'PerfValue'})) {
        $Objects{NagiosPlugin}->add_perfdata(
            label => "value",
            value => $Param{Data}[0]->{PerfValue},
            #uom => $Param{PerfUOM},
            threshold => $thresholds,
        );
    }

    $Objects{NagiosPlugin}->nagios_exit(
        return_code => $Param{Data}[0]->{'ExitCode'},
        message     => $Param{Data}[0]->{'Output'},
    );
}

sub _NagiosSubmitPassiveData {
    my %Param = @_;
    #my $CheckResultsFile = File::Temp->new(
    #    UNLINK => 0, TEMPLATE => "cXXXXXX", DIR => $CheckResultsDir
    #);

    #Daniel Ziegler openITCOCKPIT V3 patch
    my($CMDFILE) =   "/opt/openitc/nagios/var/rw/oitc.cmd"; # Nagios CMD file

    if(! -e $CMDFILE){
        #nagios.cmd does not exists - nagios not running?
        print "Unknown - File /opt/openitc/nagios/var/rw/oitc.cmd does not exists";
        exit 3;
    }

    open(my $fh, '>', $CMDFILE);

    foreach my $PassiveElement (@{$Param{Data}}) {

        my $warning = $PassiveElement->{'Warning'} // '';
        my $critical = $PassiveElement->{'Critical'} // '';

        if (defined($PassiveElement->{'PerfValue'})) {
            print $fh "[".time()."] PROCESS_SERVICE_CHECK_RESULT;".$Objects{NagiosPlugin}->opts->hostname.";".$PassiveElement->{'ServiceName'}.";".$PassiveElement->{'ExitCode'}.";".$PassiveElement->{'Output'}." | value=".$PassiveElement->{'PerfValue'} . ";" . $warning . ";" . $critical . "\n";
        }else{
            print $fh "[".time()."] PROCESS_SERVICE_CHECK_RESULT;".$Objects{NagiosPlugin}->opts->hostname.";".$PassiveElement->{'ServiceName'}.";".$PassiveElement->{'ExitCode'}.";".$PassiveElement->{'Output'}."\n";
        }

    };

    close $fh;


    print "OK - Submit passive data";
    exit STATE_OK;
}

=item _CheckServiceNameMapping
map the default service description to a description of your choice

    $RowNum->{ServiceName} = _CheckServiceNameMapping( $RowNum->{ServiceName} );

Return:
    $ServiceName = SERVICE_NAME;

=cut

################################
# helper methods               #
################################

sub _CheckServiceNameMapping {
    my $MapName = $_[0];
    foreach my $ConfigElement (keys %{$Objects{Config}->param(-block=>'Service-Mappings')}) {
        if ($MapName =~ $ConfigElement) {
            $MapName = $Objects{Config}->param(-block=>'Service-Mappings')->{$ConfigElement};
        }
    }
    return $MapName;
}

#sub _CheckServiceNameMapping {
#    my  @Param = @_;
#    my $ServiceName = $Param[0];
#
#    if (defined($Objects{Config}->param('Service-Mappings.' . $ServiceName))) {
#        $ServiceName = $Objects{Config}->param('Service-Mappings.' . $ServiceName);
#    }
#
#    return $ServiceName;
#}
=back
